
clear
clc

load 'CPIQ.txt'
data = 400*diff(log(CPIQ));
%load 'PCE.txt'
%data = 400*diff(log(PCE));



s = 3;
r = 1;
horizon = 8;
a = -2; b = 10; %% a < tau_t < b


y = data((r+1):end);
T = length(y);
M = min([T-1,100]);
y0 = data(1:r);
ypost = y((T-s+1):T)';
Xb = kron(speye(T),ones(1,r)); Xbidx = logical(Xb);
Xf = kron(speye(T),ones(1,s)); Xfidx = logical(Xf);

%% prior hyper parameters

tau0 = 0; invVtau = 1/5;
invVphi = eye(s); invVfhi = eye(r); 
nutau0 = 10; Stau0 = .02*(nutau0-1);

phih0 = .9; Vphih = 0.1^2;
muh0 = 0; Vmuh = 10;
nuh0 = 10; Sh0 = .2^2*(nuh0-1);
nuphi = 10; nufhi = 10;


if s==1
    Sphi = .001*(nuphi-1);
else
    Sphi = .001*(nuphi-s-1)*eye(s);
end

if r==1 
    Sfhi = .001*(nufhi-1);
else
    Sfhi = .001*(nufhi-r-1)*eye(r);
end

%% initialization

Sigmaphi = 0.001*eye(s);
Sigmafhi = 0.001*eye(r);
eplus = spalloc(M,1,1);
Hboldr = spdiags( -ones( T * r, 1 ), -r, speye( T * r ) );
Hbolds = spdiags( -ones( T * s, 1 ), -s, speye( T * s ) );

accept = 0;
while accept == 0
    phivec = 0.1*randn(T*s,1);
    maxroot = rootsfun_mex(phivec,s);
    if maxroot < 1
        accept = 1;
    end
end

accept = 0;
while accept == 0
    fhivec = 0.1*randn(T*r,1);
    maxroot = rootsfun_mex(fhivec,r);
    if maxroot < 1
        accept = 1;
    end
end
 
phimat = reshape(phivec',s,T)';
fhimat = reshape(fhivec',r,T)';
[P , my] = Pmy(fhimat,phimat,ypost,y0,T,r,s);
forecasts = zeros(1,horizon);

sigtau = .01;
sqrtsigtau = sqrt(sigtau);
sigh2 = .05;
phih = .9;
muh = 1;
h = log(var(y)*.8)*ones(T,1);
H = spdiags( -ones( T,1), -1,speye(T));
invS_tau = sparse(1:T,1:T,[invVtau (1/sigtau)*ones(1,T-1)],T,T);
invS_y = sparse(1:T,1:T,exp(-h));
invV_tau = H'*invS_tau*H + P'*invS_y*P;
cholH = chol(invV_tau,'lower');
taut = cholH'\(cholH\(  P'*invS_y*(P*y-my)  ));
taut = min(taut,b-.05);    taut = max(taut,a+.05);
newnutau = (T-1)/2 + nutau0;
newnuh = T/2 + nuh0;
rand('state', sum(100*clock) ); randn('state', sum(200*clock) );

accept = 0;
while accept == 0
    tau = taut + cholH'\randn(T,1);
    if  max(tau)<=b && min(tau)>=a
        accept = 1;
    end   
end



%% MCMC starts %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 N = 50000;
 
 TAU = zeros(T,N); PHI = zeros(s*T,N); FHI = zeros(r*T,N); EPLUS = zeros(M,N);
 HA = zeros(T,N); FORE = zeros(N,horizon); PAR = zeros(N,r+s+4); detP = zeros(N,1);

tic 
for idx_i = 1:N

    %% sample tau
    
    invS_tau = sparse(1:T,1:T,[invVtau (1/sigtau)*ones(1,T-1)],T,T);
    invS_y = sparse(1:T,1:T,exp(-h));
    invV_tau = H'*invS_tau*H + P'*invS_y*P;
    cholH = chol(invV_tau,'lower');
    taut = cholH'\(cholH\(  P'*invS_y*(P*y-my)  ));
    taut = min(taut,b-.05);    taut = max(taut,a+.05);
 
    % AR step
    taustar = taut;    
    u = P*(y-taustar)-my;
    logc =  -.5*exp(-h)'*u.^2 -.5*(taustar(1)-tau0)^2*invVtau + ...
        -.5*sum((taustar(2:end)- taustar(1:end-1)).^2)/sigtau + ...
        -sum(log(normcdf((b-taustar(1:end-1))/sqrtsigtau) + ...
        -normcdf((a-taustar(1:end-1))/sqrtsigtau))) + log(3);
    flag = 0; count = 0;
    while flag == 0 && count < 1000
        tauc = taut + cholH'\randn(T,1);
        if  max(tauc) <= b - .01 && min(tauc) >= a + .01
            vc = tauc-taut; uc = P*(y-tauc)-my;
            alpARc =  -.5*exp(-h)'*uc.^2 -.5*(tauc(1)-tau0)^2*invVtau +...
                -.5*sum((tauc(2:end)- tauc(1:end-1)).^2)/sigtau + ...
                -sum(log(normcdf((b-tauc(1:end-1))/sqrtsigtau) + ...
                -normcdf((a-tauc(1:end-1))/sqrtsigtau))) + .5*vc'*invV_tau*vc - logc;            
            if alpARc > log(rand)
                flag = 1;
            end
        end
        count = count + 1;
    end
    % MH-step
    if flag == 1;
        v = tau-taut;   u = P*(y-tau)-my;
        alpAR =  -.5*exp(-h)'*u.^2 -.5*(tau(1)-tau0)^2*invVtau + ...
            -.5*sum((tau(2:end)- tau(1:end-1)).^2)/sigtau + ...
            - sum(log(normcdf((b-tau(1:end-1))/sqrtsigtau) + ...
            -normcdf((a-tau(1:end-1))/sqrtsigtau))) + .5*v'*invV_tau*v - logc;
        if alpAR < 0
            alpMH = 1;
        elseif alpARc < 0
            alpMH = - alpAR;
        else
            alpMH = alpARc - alpAR;
        end        
        if alpMH > log(rand)
            tau = tauc;
        end     
    end
    
    ystar = y - tau;

    %% sample fhi    

    Xlag = lagmatrix([y0;ystar;ypost'],[0:max(r,s)]);
    U = zeros(T,r+1);
    for j = 1:(r+1)
        uTmp = lagmatrix(Xlag(:,j),[0:s]);
        uTmp = uTmp((r+s+1):end,(s+1):-1:1);
        U(:,j) = sum([ones(T,1), -phimat].*uTmp,2);
    end
    u = U(:,1);
    Xlagt = U(:,2:(r+1))';
    Xb(Xbidx) = Xlagt(:);
    
    invOmegafhi = [invVfhi sparse(r,r*(T-1)); ...
        sparse(r*(T-1),r) kron(speye(T-1),Sigmafhi\speye(r))];
    HinvOH = Hboldr'*invOmegafhi*Hboldr;
    XS = Xb'*invS_y;
    invV_fhi = HinvOH + XS*Xb;
    C_fhi = chol(invV_fhi,'lower');
    fhihat = C_fhi'\(C_fhi\(  XS*u  ));
    
    flag = 0; count = 0;
    while flag == 0 && count < 1000
        fhivecstar = fhihat + C_fhi'\randn(r*T,1);
        flag  = rootsfunn_mex(fhivecstar,r);
        count = count + 1;
    end

    fhimatstar = reshape(fhivecstar',r,T)';
    [Pstar , mystar] = Pmy(fhimatstar,phimat,ypost,y0,T,r,s);
    
    if flag == 1 && abs(det(Pstar)) > 0
        alpMH = log(abs(det(Pstar))) - log(abs(det(P)));
        if alpMH > log(rand)
            fhivec = fhivecstar;
            P = Pstar;
            my = mystar;
        end
    end
    
    fhimat = reshape(fhivec',r,T)';
    [F, gamma_y0] = Fmat(fhivec,y0,T,r);
    %max(abs((P*ystar - my) - (u-Xb*fhivec)))
    
    
    %% sample phi

    Xlead = Xlag((s+1):size(Xlag,1),(s+1):-1:1);
    V = zeros(T,s+1);
    for j = 1:(s+1)
        vTmp = lagmatrix(Xlead(:,j),[0:r]);
        V(:,j) = sum([ones(T,1),-fhimat].*vTmp((r+1):size(Xlead,1),:),2);
    end
    v = V(:,1);
    Xleadt = V(:,2:(s+1))';
    Xf(Xfidx) = Xleadt(:);
    
    invOmegaphi = [invVphi sparse(s,s*(T-1)); ...
        sparse(s*(T-1),s) kron(speye(T-1),Sigmaphi\speye(s))];
    HinvOH = Hbolds'*invOmegaphi*Hbolds;
    XS = Xf'*invS_y;
    invV_phi = HinvOH + XS*Xf;
    C_phi = chol(invV_phi,'lower');
    phihat = C_phi'\(C_phi\(  XS*v  ));
    
    flag = 0; count = 0;
    while flag == 0 && count < 1000
        phivecstar = phihat + C_phi'\randn(s*T,1);
        flag  = rootsfunn_mex(phivecstar,s);
        count = count + 1;
    end

    phimatstar = reshape(phivecstar',s,T)';
    [Pstar , mystar] = Pmy(fhimat,phimatstar,ypost,y0,T,r,s);
    
    if flag == 1 && abs(det(Pstar)) > 0
        alpMH = log(abs(det(Pstar))) - log(abs(det(P)));
        if alpMH > log(rand)
            phivec = phivecstar;
            P = Pstar;
            my = mystar;
        end
    end
    
    phimat = reshape(phivec',s,T)';
    [X_beta, B] = betafun_mex(phivec,s);
    X_beta = sparse(X_beta(:,1:M));
    B = sparse(B);
    invB = B\speye(T);
    ytilde = P*ystar;
    %max(abs((P*ystar - my) - (v-Xf*phivec)))
    
        
    %% sample h    
    
    Ystar = log((ytilde-my).^2 + .0001);
    h = SV(Ystar,h,phih,sigh2,muh);
    invS_y = sparse(1:T,1:T,exp(-h));
    
    
    %% sample eplus
    
    Hphih = speye(M) - sparse(2:M,1:(M-1),phih*ones(1,M-1),M,M);
    HinvSH = Hphih'*sparse(1:M,1:M, 1/sigh2*ones(M,1))*Hphih;
    hplushat = Hphih\[muh*(1-phih) + phih*h(T); muh*(1-phih)*ones(M-1,1)];
    C_hplus = chol(HinvSH,'lower');
    hplus = hplushat + C_hplus'\randn(M,1);
    
    BXSB = (invB*X_beta)'*invS_y*invB;
    invV_eplus = sparse(1:M,1:M,exp(-hplus)) + BXSB*X_beta;
    C_eplus = chol(invV_eplus,'lower');
    eplushat = C_eplus'\(C_eplus\(  BXSB*(F*ystar-gamma_y0)  ));
    
    phi = phimat(T,:); fhi = fhimat(T,:);
    [postfhi,  postphi] = postcoeffunn(phi,fhi,Sigmaphi,Sigmafhi,horizon);  
    
    % AR-step:
    forecastshat = forecastfun_mex(postphi,postfhi,eplushat,horizon,ystar);
    yposthat = forecastshat(1:s);
    [~ , myhat] = Pmy(fhimat,phimat,yposthat,y0,T,r,s);
    logc = log(3) - 0.5*sum(exp(-h((T-s+1):T)).*(ytilde((T-s+1):T)-myhat((T-s+1):T)).^2) + ...
        - 0.5*exp(-hplus)'*eplushat.^2;
    accept = 0; count = 0;
    while accept == 0 && count < 1000
        eplusstar = eplushat + sqrt(5/chi2rnd(5))*(C_eplus'\randn(M,1));
        forecastsstar = forecastfun_mex(postphi,postfhi,eplusstar,horizon,ystar);
        ypoststar = forecastsstar(1:s);
        [~ , mystar] = Pmy(fhimat,phimat,ypoststar,y0,T,r,s);
        ceplus = eplusstar - eplushat;
        alpARc = - logc - 0.5*exp(-hplus)'*eplusstar.^2 + ...
            - 0.5*sum(exp(-h((T-s+1):T)).*(ytilde((T-s+1):T)-mystar((T-s+1):T)).^2) + ...
           + 0.5*(M+5)*log(1+ceplus'*invV_eplus*ceplus/5); % + 0.5*ceplus'*invV_eplus*ceplus;
        if alpARc > log(rand)
            accept = 1;
        end
        count = count + 1;
    end
    % MH-step
    if accept == 1
        veplus = eplus - eplushat;
        alpAR = - logc - 0.5*exp(-hplus)'*eplus.^2 + ...
            - 0.5*sum(exp(-h((T-s+1):T)).*(ytilde((T-s+1):T)-my((T-s+1):T)).^2) + ...
          + 0.5*(M+5)*log(1+veplus'*invV_eplus*veplus/5); % + 0.5*veplus'*invV_eplus*veplus;
        if alpAR < 0
            alpMH = 1;
        elseif alpARc < 0
            alpMH = - alpAR;
        else
            alpMH = alpARc - alpAR;
        end
        if alpMH > log(rand)
            eplus = eplusstar;
            ypost = ypoststar;
            forecasts = forecastsstar;
            my = mystar;
        end
    end


    %% sample sigtau
 
    errsigtau = tau(2:end) - tau(1:end-1);
    newStau = Stau0 + sum(errsigtau.^2)/2;
    sigtauc = 1/gamrnd(newnutau, 1/newStau);    
    sqrtsigtauc = sqrt(sigtauc);
    siga = a - tau(1:end-1); 
    sigb = b - tau(1:end-1);    
    alpMH = -sum(log(normcdf(sigb/sqrtsigtauc)-normcdf(siga/sqrtsigtauc))) + ...
        sum(log(normcdf(sigb/sqrtsigtau)-normcdf(siga/sqrtsigtau)));
    if alpMH > log(rand)
        sigtau = sigtauc;         
        sqrtsigtau = sqrtsigtauc;
    end


    %% sample sigh2
    
    %hstore = h;
    %h = [h;hplus];
    errh = [(h(1)-muh)*sqrt(1-phih^2); h(2:end)-phih*h(1:end-1)-muh*(1-phih)];    
    newSh = Sh0 + sum(errh.^2)/2;    
    sigh2 = 1/gamrnd(newnuh, 1/newSh);   
    
    
    %% sample muh and phih jointly
        
    sum1 = sum(h(2:end));
    sum2 = sum(h(1:end-1));    
    flam = @(x) -(1-x(2))^2/(2*sigh2)*(h(1)-x(1))^2 - 1/(2*sigh2) * ...
        sum((h(2:end) - x(2)*h(1:end-1) - x(1)*(1-x(2))).^2)...
        - 1/(2*Vphih)*(x(2)-phih0)^2 - 1/(2*Vmuh)*(x(1)-muh0)^2;
    [lamc g] = proplam(h,sigh2);
    MHprob = flam(lamc) - flam([muh; phih]) + g([muh; phih]) - g(lamc);
    if abs(lamc(2)) < 1
        if exp(MHprob) > rand
            muh = lamc(1);
            phih = lamc(2);
        end
    end
    %h = hstore;

    
    %% sample Sigmafhi
    
    if r > 1
        err = fhimat(2:end,:) - fhimat(1:end-1,:);
        Sigmafhi = iwishrnd(Sfhi + err'*err, nufhi + T-1);
    else
        err = fhivec(2:end) - fhivec(1:end-1);
        newSfhi = Sfhi + sum(err.^2)/2;
        Sigmafhi = 1/gamrnd((T-1)/2+nufhi,1/newSfhi);
    end

    
    %% sample Sigmaphi
    if s > 1
        err = phimat(2:end,:) - phimat(1:end-1,:);
        Sigmaphi = iwishrnd(Sphi + err'*err, nuphi + T-1);
    else
        err = phivec(2:end) - phivec(1:end-1);
        newSphi = Sphi + sum(err.^2)/2;
        Sigmaphi = 1/gamrnd((T-1)/2+nuphi,1/newSphi);
    end

 
    %%
    
    e_tau = zeros(1,horizon+1); e_tau(1) = tau(T);
    for t = 2:horizon+1
        accept = 0; count = 0;
        while accept == 0 && count < 100
            e_tau(t) = e_tau(t-1) + sqrtsigtau*randn(1);
        if  e_tau(t) < b && e_tau(t) > a
            accept = 1;
        end
        count = count + 1;
        end
        if accept == 0
            if e_tau(t) >= b e_tau(t) = b; end
            if e_tau(t) <= a e_tau(t) = a; end
        end
    end
    e_tau = e_tau(2:end);
    
    TAU(:,idx_i) = tau; PHI(:,idx_i) = phimat(:); FHI(:,idx_i) = fhimat(:);
    EPLUS(:,idx_i) = eplus; HA(:,idx_i) = h; FORE(idx_i,:) = forecasts + e_tau;
    PAR(idx_i,:) = [muh,phih,sigh2,sigtau,diag(Sigmaphi)',diag(Sigmafhi)'];
    detP(idx_i) = det(Pstar);
    idx_i

end


mean(diff(PHI(10,:))==0)
mean(diff(FHI(10,:))==0)
mean(diff(TAU(10,:))==0)
mean(diff(EPLUS(10,:))==0)
toc

plot(quantile(TAU',0.84))
plot(EPLUS(1,:))
plot(FORE(:,1))

